/*
 * Copyright (c) 2014 Bartek Fabiszewski
 * http://www.fabiszewski.net
 *
 * This file is part of libmobi.
 * Licensed under LGPL, either version 3, or any later.
 * See <http://www.gnu.org/licenses/>
 */

#ifndef libmobi_write_h
#define libmobi_write_h

#include "config.h"
#include "mobi.h"
#include "buffer.h"


#endif
