/** @file meta.h
 *
 * Copyright (c) 2016 Bartek Fabiszewski
 * http://www.fabiszewski.net
 *
 * This file is part of libmobi.
 * Licensed under LGPL, either version 3, or any later.
 * See <http://www.gnu.org/licenses/>
 */

#ifndef libmobi_meta_h
#define libmobi_meta_h

#include "config.h"
#include "mobi.h"

#endif /* libmobi_meta_h */
